@@ -1,5 +1,5 @@
 /*
- * Copyright 2019-2020 OARC, Inc.
+ * Copyright 2019-2021 OARC, Inc.
  * Copyright 2017-2018 Akamai Technologies
  * Copyright 2006-2016 Nominum, Inc.
  * All rights reserved.
@@ -23,6 +23,17 @@
 
 #include "config.h"
 
+#include "datafile.h"
+#include "dns.h"
+#include "log.h"
+#include "net.h"
+#include "opt.h"
+#include "util.h"
+#include "os.h"
+#include "list.h"
+#include "result.h"
+#include "buffer.h"
+
 #include <errno.h>
 #include <stdbool.h>
 #include <stdio.h>
@@ -35,55 +46,40 @@
 #include <openssl/err.h>
 #include <signal.h>
 
-#include <isc/buffer.h>
-#include <isc/file.h>
-#include <isc/list.h>
-#include <isc/mem.h>
-#include <isc/print.h>
-#include <isc/region.h>
-#include <isc/result.h>
-#include <isc/sockaddr.h>
-#include <isc/types.h>
-
-#include <dns/result.h>
-
-#include "datafile.h"
-#include "dns.h"
-#include "log.h"
-#include "net.h"
-#include "opt.h"
-#include "util.h"
-#include "os.h"
-
 /*
  * Global stuff
  */
 
 #define DEFAULT_SERVER_NAME "127.0.0.1"
 #define DEFAULT_SERVER_PORT 53
-#define DEFAULT_SERVER_TLS_PORT 853
-#define DEFAULT_SERVER_PORTS "udp/tcp 53 or tls 853"
+#define DEFAULT_SERVER_DOT_PORT 853
+#define DEFAULT_SERVER_PORTS "udp/tcp 53 or DoT 853"
 #define DEFAULT_LOCAL_PORT 0
 #define DEFAULT_SOCKET_BUFFER 32
 #define DEFAULT_TIMEOUT 45
 #define DEFAULT_MAX_OUTSTANDING (64 * 1024)
+#define DEFAULT_MAX_FALL_BEHIND 1000
 
-#define MAX_INPUT_DATA (4 * 1024)
+#define MAX_INPUT_DATA (64 * 1024)
 
 #define TIMEOUT_CHECK_TIME 5000000
 
+#define DNS_RCODE_NOERROR 0
+#define DNS_RCODE_NXDOMAIN 3
+
 struct query_info;
 
-typedef ISC_LIST(struct query_info) query_list;
+typedef perf_list(struct query_info) query_list;
 
 typedef struct query_info {
     uint64_t sent_timestamp;
+    bool     is_inprogress;
+
     /*
      * This link links the query into the list of outstanding
      * queries or the list of available query IDs.
      */
-    ISC_LINK(struct query_info)
-    link;
+    perf_link(struct query_info);
     /*
      * The list this query is on.
      */
@@ -95,13 +91,11 @@ static query_list instanding_list;
 
 static query_info* queries;
 
-static isc_mem_t* mctx;
-
-static isc_sockaddr_t          server_addr;
-static isc_sockaddr_t          local_addr;
-static unsigned int            nsocks;
-static struct perf_net_socket* socks;
-static enum perf_net_mode      mode;
+static perf_sockaddr_t          server_addr;
+static perf_sockaddr_t          local_addr;
+static unsigned int             nsocks;
+static struct perf_net_socket** socks;
+static enum perf_net_mode       mode;
 
 static int dummypipe[2];
 
@@ -152,6 +146,7 @@ static uint64_t num_queries_outstanding;
 static uint64_t num_responses_received;
 static uint64_t num_queries_timed_out;
 static uint64_t rcodecounts[16];
+static uint64_t num_reconnections;
 
 static uint64_t time_now;
 static uint64_t time_of_program_start;
@@ -173,6 +168,9 @@ typedef struct {
     int    responses;
     int    failures;
     double latency_sum;
+
+    int    connections;
+    double conn_latency_sum;
 } ramp_bucket;
 
 /* Pointer to array of n_buckets ramp_bucket structures */
@@ -199,9 +197,12 @@ static enum phase phase = PHASE_RAMP;
 /* The time when the sustain/wait phase began */
 static uint64_t sustain_phase_began, wait_phase_began;
 
-static perf_dnstsigkey_t* tsigkey;
+static perf_tsigkey_t* tsigkey;
 
-static bool verbose;
+static bool         verbose;
+static unsigned int max_fall_behind;
+
+const char* progname = "resperf";
 
 static char*
 stringify(double value, int precision)
@@ -212,6 +213,9 @@ stringify(double value, int precision)
     return buf;
 }
 
+static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t event, uint64_t elapsed_time);
+static void perf__net_sent(struct perf_net_socket* sock, uint16_t qid);
+
 static void
 setup(int argc, char** argv)
 {
@@ -227,17 +231,6 @@ setup(int argc, char** argv)
     unsigned int i;
     const char*  _mode = 0;
 
-#ifdef HAVE_ISC_MEM_CREATE_RESULT
-    isc_result_t result = isc_mem_create(0, 0, &mctx);
-    if (result != ISC_R_SUCCESS)
-        perf_log_fatal("creating memory context: %s",
-            isc_result_totext(result));
-#else
-    isc_mem_create(&mctx);
-#endif
-
-    dns_result_register();
-
     sock_family     = AF_UNSPEC;
     server_port     = 0;
     local_port      = DEFAULT_LOCAL_PORT;
@@ -250,11 +243,12 @@ setup(int argc, char** argv)
     nsocks          = 1;
     mode            = sock_udp;
     verbose         = false;
+    max_fall_behind = DEFAULT_MAX_FALL_BEHIND;
 
     perf_opt_add('f', perf_opt_string, "family",
         "address family of DNS transport, inet or inet6", "any",
         &family);
-    perf_opt_add('M', perf_opt_string, "mode", "set transport mode: udp, tcp or tls", "udp", &_mode);
+    perf_opt_add('M', perf_opt_string, "mode", "set transport mode: udp, tcp or dot", "udp", &_mode);
     perf_opt_add('s', perf_opt_string, "server_addr",
         "the server to query", DEFAULT_SERVER_NAME, &server_name);
     perf_opt_add('p', perf_opt_port, "port",
@@ -298,21 +292,31 @@ setup(int argc, char** argv)
         "the maximum acceptable query loss, in percent",
         stringify(max_loss_percent, 0), &max_loss_percent);
     perf_opt_add('C', perf_opt_uint, "clients",
-        "the number of clients to act as", NULL, &nsocks);
+        "the number of clients to act as", stringify(1, 0), &nsocks);
     perf_opt_add('q', perf_opt_uint, "num_outstanding",
         "the maximum number of queries outstanding",
         stringify(DEFAULT_MAX_OUTSTANDING, 0), &max_outstanding);
     perf_opt_add('v', perf_opt_boolean, NULL,
         "verbose: report additional information to stdout",
         NULL, &verbose);
+    bool log_stdout = false;
+    perf_opt_add('W', perf_opt_boolean, NULL, "log warnings and errors to stdout instead of stderr", NULL, &log_stdout);
+    bool reopen_datafile = false;
+    perf_opt_add('R', perf_opt_boolean, NULL, "reopen datafile on end, allow for infinit use of it", NULL, &reopen_datafile);
+    perf_opt_add('F', perf_opt_zpint, "fall_behind", "the maximum number of queries that is allowed to fall behind, zero to disable",
+        stringify(DEFAULT_MAX_FALL_BEHIND, 0), &max_fall_behind);
 
     perf_opt_parse(argc, argv);
 
+    if (log_stdout) {
+        perf_log_tostdout();
+    }
+
     if (_mode != 0)
         mode = perf_net_parsemode(_mode);
 
     if (!server_port) {
-        server_port = mode == sock_tls ? DEFAULT_SERVER_TLS_PORT : DEFAULT_SERVER_PORT;
+        server_port = mode == sock_dot ? DEFAULT_SERVER_DOT_PORT : DEFAULT_SERVER_PORT;
     }
 
     if (max_outstanding > nsocks * DEFAULT_MAX_OUTSTANDING)
@@ -324,36 +328,46 @@ setup(int argc, char** argv)
         perf_log_fatal("rampup_time and constant_traffic_time must not "
                        "both be 0");
 
-    ISC_LIST_INIT(outstanding_list);
-    ISC_LIST_INIT(instanding_list);
-    queries = isc_mem_get(mctx, max_outstanding * sizeof(query_info));
-    if (queries == NULL)
+    perf_list_init(outstanding_list);
+    perf_list_init(instanding_list);
+    if (!(queries = calloc(max_outstanding, sizeof(query_info)))) {
         perf_log_fatal("out of memory");
+    }
     for (i = 0; i < max_outstanding; i++) {
-        ISC_LINK_INIT(&queries[i], link);
-        ISC_LIST_APPEND(instanding_list, &queries[i], link);
+        perf_link_init(&queries[i]);
+        perf_list_append(instanding_list, &queries[i]);
         queries[i].list = &instanding_list;
     }
 
     if (family != NULL)
         sock_family = perf_net_parsefamily(family);
     perf_net_parseserver(sock_family, server_name, server_port, &server_addr);
-    perf_net_parselocal(isc_sockaddr_pf(&server_addr), local_name,
+    perf_net_parselocal(server_addr.sa.sa.sa_family, local_name,
         local_port, &local_addr);
 
-    input = perf_datafile_open(mctx, filename);
+    input = perf_datafile_open(filename);
+    if (reopen_datafile) {
+        perf_datafile_setmaxruns(input, -1);
+    }
 
     if (dnssec)
         edns = true;
 
     if (tsigkey_str != NULL)
-        tsigkey = perf_dns_parsetsigkey(tsigkey_str, mctx);
+        tsigkey = perf_tsig_parsekey(tsigkey_str);
 
-    socks = isc_mem_get(mctx, nsocks * sizeof(*socks));
-    if (socks == NULL)
+    if (!(socks = calloc(nsocks, sizeof(*socks)))) {
         perf_log_fatal("out of memory");
-    for (i = 0; i < nsocks; i++)
+    }
+    for (i = 0; i < nsocks; i++) {
         socks[i] = perf_net_opensocket(mode, &server_addr, &local_addr, i, bufsize);
+        if (!socks[i]) {
+            perf_log_fatal("perf_net_opensocket(): no socket returned, out of memory?");
+        }
+        socks[i]->data  = (void*)(intptr_t)i;
+        socks[i]->sent  = perf__net_sent;
+        socks[i]->event = perf__net_event;
+    }
 }
 
 static void
@@ -363,11 +377,7 @@ cleanup(void)
 
     perf_datafile_close(&input);
     for (i = 0; i < nsocks; i++)
-        (void)perf_net_close(&socks[i]);
-    isc_mem_put(mctx, socks, nsocks * sizeof(*socks));
-    isc_mem_put(mctx, queries, max_outstanding * sizeof(query_info));
-    isc_mem_put(mctx, buckets, n_buckets * sizeof(ramp_bucket));
-    isc_mem_destroy(&mctx);
+        (void)perf_net_close(socks[i]);
     close(dummypipe[0]);
     close(dummypipe[1]);
 }
@@ -390,6 +400,30 @@ find_bucket(uint64_t when)
     return &buckets[i];
 }
 
+static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t event, uint64_t elapsed_time)
+{
+    ramp_bucket* b = find_bucket(time_now);
+
+    switch (event) {
+    case perf_socket_event_reconnect:
+        num_reconnections++;
+    case perf_socket_event_connect:
+        b->connections++;
+        b->conn_latency_sum += elapsed_time / (double)MILLION;
+    }
+}
+
+static void perf__net_sent(struct perf_net_socket* sock, uint16_t qid)
+{
+    ramp_bucket* b = find_bucket(time_now);
+
+    b->queries++;
+
+    size_t idx = (size_t)qid * nsocks + (intptr_t)sock->data;
+    assert(idx < max_outstanding);
+    queries[idx].sent_timestamp = time_now;
+}
+
 /*
  * print_statistics:
  *   Print out statistics based on the results of the test
@@ -425,6 +459,7 @@ print_statistics(void)
             (rcodecounts[i] * 100.0) / num_responses_received);
     }
     printf("\n");
+    printf("  Reconnection(s):      %" PRIu64 "\n", num_reconnections);
     printf("  Run time (s):         %u.%06u\n",
         (unsigned int)(run_time / MILLION),
         (unsigned int)(run_time % MILLION));
@@ -454,9 +489,10 @@ init_buckets(int n)
     ramp_bucket* p;
     int          i;
 
-    p = isc_mem_get(mctx, n * sizeof(*p));
-    if (p == NULL)
+    if (!(p = calloc(n, sizeof(*p)))) {
         perf_log_fatal("out of memory");
+        return 0; // fix clang scan-build
+    }
     for (i = 0; i < n; i++) {
         p[i].queries = p[i].responses = p[i].failures = 0;
         p[i].latency_sum                              = 0.0;
@@ -466,33 +502,27 @@ init_buckets(int n)
 
 /*
  * Send a query based on a line of input.
- * Return ISC_R_NOMORE if we ran out of query IDs.
+ * Return PERF_R_NOMORE if we ran out of query IDs.
  */
-static isc_result_t
-do_one_line(isc_buffer_t* lines, isc_buffer_t* msg)
+static perf_result_t
+do_one_line(perf_buffer_t* lines, perf_buffer_t* msg)
 {
     query_info*    q;
     unsigned int   qid;
     unsigned int   sock;
-    isc_region_t   used;
+    perf_region_t  used;
     unsigned char* base;
     unsigned int   length;
-    isc_result_t   result;
+    perf_result_t  result;
 
-    isc_buffer_clear(lines);
-    result = perf_datafile_next(input, lines, false);
-    if (result != ISC_R_SUCCESS)
-        perf_log_fatal("ran out of query data");
-    isc_buffer_usedregion(lines, &used);
-
-    q = ISC_LIST_HEAD(instanding_list);
+    q = perf_list_head(instanding_list);
     if (!q)
-        return (ISC_R_NOMORE);
+        return (PERF_R_NOMORE);
     qid  = (q - queries) / nsocks;
     sock = (q - queries) % nsocks;
 
-    if (socks[sock].sending) {
-        if (perf_net_sockready(&socks[sock], dummypipe[0], TIMEOUT_CHECK_TIME) == -1) {
+    while (q->is_inprogress) {
+        if (perf_net_sockready(socks[sock], dummypipe[0], TIMEOUT_CHECK_TIME) == -1) {
             if (errno == EINPROGRESS) {
                 if (verbose) {
                     perf_log_warning("network congested, packet sending in progress");
@@ -500,73 +530,89 @@ do_one_line(isc_buffer_t* lines, isc_buffer_t* msg)
             } else {
                 if (verbose) {
                     char __s[256];
-                    perf_log_warning("failed to send packet: %s", perf_strerror_r(errno, __s, sizeof(__s)));
+                    perf_log_warning("failed to check socket readiness: %s", perf_strerror_r(errno, __s, sizeof(__s)));
                 }
             }
-            return (ISC_R_FAILURE);
+            return (PERF_R_FAILURE);
         }
 
-        ISC_LIST_UNLINK(instanding_list, q, link);
-        ISC_LIST_PREPEND(outstanding_list, q, link);
+        q->is_inprogress = false;
+        perf_list_unlink(instanding_list, q);
+        perf_list_prepend(outstanding_list, q);
         q->list = &outstanding_list;
 
         num_queries_sent++;
         num_queries_outstanding++;
 
-        q = ISC_LIST_HEAD(instanding_list);
+        q = perf_list_head(instanding_list);
         if (!q)
-            return (ISC_R_NOMORE);
+            return (PERF_R_NOMORE);
         qid  = (q - queries) / nsocks;
         sock = (q - queries) % nsocks;
     }
 
-    isc_buffer_clear(msg);
-    result = perf_dns_buildrequest(NULL, (isc_textregion_t*)&used,
-        qid, edns, dnssec, tsigkey, NULL, msg);
-    if (result != ISC_R_SUCCESS)
-        return (result);
-
-    q->sent_timestamp = time_now;
-
-    switch (perf_net_sockready(&socks[sock], dummypipe[0], TIMEOUT_CHECK_TIME)) {
+    switch (perf_net_sockready(socks[sock], dummypipe[0], TIMEOUT_CHECK_TIME)) {
     case 0:
         if (verbose) {
             perf_log_warning("failed to send packet: socket %d not ready", sock);
         }
-        return (ISC_R_FAILURE);
+        return (PERF_R_FAILURE);
     case -1:
-        perf_log_warning("failed to send packet: socket %d readiness check timed out", sock);
-        return (ISC_R_FAILURE);
+        if (errno == EINPROGRESS) {
+            if (verbose) {
+                perf_log_warning("network congested, packet sending in progress");
+            }
+        } else {
+            perf_log_warning("failed to send packet: socket %d not ready", sock);
+        }
+        return (PERF_R_FAILURE);
     default:
         break;
     }
 
-    base   = isc_buffer_base(msg);
-    length = isc_buffer_usedlength(msg);
-    if (perf_net_sendto(&socks[sock], base, length, 0,
-            &server_addr.type.sa, server_addr.length)
+    perf_buffer_clear(lines);
+    result = perf_datafile_next(input, lines, false);
+    if (result != PERF_R_SUCCESS)
+        perf_log_fatal("ran out of query data");
+    perf_buffer_usedregion(lines, &used);
+
+    perf_buffer_clear(msg);
+    result = perf_dns_buildrequest(&used, qid,
+        edns, dnssec, false,
+        tsigkey, 0,
+        msg);
+    if (result != PERF_R_SUCCESS)
+        return (result);
+
+    q->sent_timestamp = time_now;
+
+    base   = perf_buffer_base(msg);
+    length = perf_buffer_usedlength(msg);
+    if (perf_net_sendto(socks[sock], qid, base, length, 0,
+            &server_addr.sa.sa, server_addr.length)
         < 1) {
         if (errno == EINPROGRESS) {
             if (verbose) {
                 perf_log_warning("network congested, packet sending in progress");
             }
+            q->is_inprogress = true;
         } else {
             if (verbose) {
                 char __s[256];
                 perf_log_warning("failed to send packet: %s", perf_strerror_r(errno, __s, sizeof(__s)));
             }
         }
-        return (ISC_R_FAILURE);
+        return (PERF_R_FAILURE);
     }
 
-    ISC_LIST_UNLINK(instanding_list, q, link);
-    ISC_LIST_PREPEND(outstanding_list, q, link);
+    perf_list_unlink(instanding_list, q);
+    perf_list_prepend(outstanding_list, q);
     q->list = &outstanding_list;
 
     num_queries_sent++;
     num_queries_outstanding++;
 
-    return ISC_R_SUCCESS;
+    return PERF_R_SUCCESS;
 }
 
 static void
@@ -604,8 +650,17 @@ try_process_response(unsigned int sockindex)
     ramp_bucket*  b;
     int           n;
 
+    if (perf_net_sockready(socks[sockindex], dummypipe[0], TIMEOUT_CHECK_TIME) == -1) {
+        if (errno != EINPROGRESS) {
+            if (verbose) {
+                char __s[256];
+                perf_log_warning("failed to check socket readiness: %s", perf_strerror_r(errno, __s, sizeof(__s)));
+            }
+        }
+    }
+
     packet_header = (uint16_t*)packet_buffer;
-    n             = perf_net_recv(&socks[sockindex], packet_buffer, sizeof(packet_buffer), 0);
+    n             = perf_net_recv(socks[sockindex], packet_buffer, sizeof(packet_buffer), 0);
     if (n < 0) {
         if (errno == EAGAIN || errno == EINTR) {
             return;
@@ -624,14 +679,15 @@ try_process_response(unsigned int sockindex)
     qid   = ntohs(packet_header[0]);
     rcode = ntohs(packet_header[1]) & 0xF;
 
-    q = &queries[qid * nsocks + sockindex];
-    if (q->list != &outstanding_list) {
+    size_t idx = qid * nsocks + sockindex;
+    if (idx >= max_outstanding || queries[idx].list != &outstanding_list) {
         perf_log_warning("received a response with an unexpected id: %u", qid);
         return;
     }
+    q = &queries[idx];
 
-    ISC_LIST_UNLINK(outstanding_list, q, link);
-    ISC_LIST_APPEND(instanding_list, q, link);
+    perf_list_unlink(outstanding_list, q);
+    perf_list_append(instanding_list, q);
     q->list = &instanding_list;
 
     num_queries_outstanding--;
@@ -639,7 +695,7 @@ try_process_response(unsigned int sockindex)
     latency = (time_now - q->sent_timestamp) / (double)MILLION;
     b       = find_bucket(q->sent_timestamp);
     b->responses++;
-    if (!(rcode == dns_rcode_noerror || rcode == dns_rcode_nxdomain))
+    if (!(rcode == DNS_RCODE_NOERROR || rcode == DNS_RCODE_NXDOMAIN))
         b->failures++;
     b->latency_sum += latency;
     num_responses_received++;
@@ -652,11 +708,11 @@ retire_old_queries(void)
     query_info* q;
 
     while (true) {
-        q = ISC_LIST_TAIL(outstanding_list);
+        q = perf_list_tail(outstanding_list);
         if (q == NULL || (time_now - q->sent_timestamp) < query_timeout)
             break;
-        ISC_LIST_UNLINK(outstanding_list, q, link);
-        ISC_LIST_APPEND(instanding_list, q, link);
+        perf_list_unlink(outstanding_list, q);
+        perf_list_append(instanding_list, q);
         q->list = &instanding_list;
 
         num_queries_outstanding--;
@@ -674,31 +730,16 @@ num_scheduled(uint64_t time_since_start)
     }
 }
 
-static void
-handle_sigpipe(int sig)
-{
-    (void)sig;
-    switch (mode) {
-    case sock_tcp:
-    case sock_tls:
-        // if connection is closed it will generate a signal
-        perf_log_fatal("SIGPIPE received, connection(s) likely closed, can't continue");
-        break;
-    default:
-        break;
-    }
-}
-
 int main(int argc, char** argv)
 {
     int           i;
     FILE*         plotf;
-    isc_buffer_t  lines, msg;
+    perf_buffer_t lines, msg;
     char          input_data[MAX_INPUT_DATA];
     unsigned char outpacket_buffer[MAX_EDNS_PACKET];
     unsigned int  max_packet_size;
     unsigned int  current_sock;
-    isc_result_t  result;
+    perf_result_t result;
 
     printf("DNS Resolution Performance Testing Tool\n"
            "Version " PACKAGE_VERSION "\n\n");
@@ -714,12 +755,20 @@ int main(int argc, char** argv)
     if (pipe(dummypipe) < 0)
         perf_log_fatal("creating pipe");
 
-    perf_os_handlesignal(SIGPIPE, handle_sigpipe);
+    switch (mode) {
+    case sock_tcp:
+    case sock_dot:
+        // block SIGPIPE for TCP/DOT mode, if connection is closed it will generate a signal
+        perf_os_blocksignal(SIGPIPE, true);
+        break;
+    default:
+        break;
+    }
 
-    isc_buffer_init(&lines, input_data, sizeof(input_data));
+    perf_buffer_init(&lines, input_data, sizeof(input_data));
 
     max_packet_size = edns ? MAX_EDNS_PACKET : MAX_UDP_PACKET;
-    isc_buffer_init(&msg, outpacket_buffer, max_packet_size);
+    perf_buffer_init(&msg, outpacket_buffer, max_packet_size);
 
     traffic_time = ramp_time + sustain_time;
     end_time     = traffic_time + wait_time;
@@ -727,10 +776,10 @@ int main(int argc, char** argv)
     n_buckets = (traffic_time + bucket_interval - 1) / bucket_interval;
     buckets   = init_buckets(n_buckets);
 
-    time_now              = get_time();
+    time_now              = perf_get_time();
     time_of_program_start = time_now;
 
-    printf("[Status] Command line: %s", isc_file_basename(argv[0]));
+    printf("[Status] Command line: %s", progname);
     for (i = 1; i < argc; i++) {
         printf(" %s", argv[i]);
     }
@@ -738,7 +787,8 @@ int main(int argc, char** argv)
 
     printf("[Status] Sending\n");
 
-    current_sock = 0;
+    int try_responses = (max_qps / max_outstanding) + 1;
+    current_sock      = 0;
     for (;;) {
         int      should_send;
         uint64_t time_since_start = time_now - time_of_program_start;
@@ -752,13 +802,13 @@ int main(int argc, char** argv)
                 enter_wait_phase();
             break;
         case PHASE_WAIT:
-            if (time_since_start >= end_time || ISC_LIST_EMPTY(outstanding_list))
+            if (time_since_start >= end_time || perf_list_empty(outstanding_list))
                 goto end_loop;
             break;
         }
         if (phase != PHASE_WAIT) {
             should_send = num_scheduled(time_since_start) - num_queries_sent;
-            if (should_send >= 1000) {
+            if (max_fall_behind && should_send >= max_fall_behind) {
                 printf("[Status] Fell behind by %d queries, "
                        "ending test at %.0f qps\n",
                     should_send, (max_qps * time_since_start) / ramp_time);
@@ -766,22 +816,25 @@ int main(int argc, char** argv)
             }
             if (should_send > 0) {
                 result = do_one_line(&lines, &msg);
-                if (result == ISC_R_SUCCESS)
+                if (result == PERF_R_SUCCESS)
                     find_bucket(time_now)->queries++;
-                if (result == ISC_R_NOMORE) {
+                if (result == PERF_R_NOMORE) {
                     printf("[Status] Reached %u outstanding queries\n",
                         max_outstanding);
                     enter_wait_phase();
                 }
             }
         }
-        try_process_response(current_sock++);
-        current_sock = current_sock % nsocks;
+        for (i = try_responses; i--;) {
+            try_process_response(current_sock++);
+            if (current_sock >= nsocks)
+                current_sock = 0;
+        }
         retire_old_queries();
-        time_now = get_time();
+        time_now = perf_get_time();
     }
 end_loop:
-    time_now           = get_time();
+    time_now           = perf_get_time();
     time_of_end_of_run = time_now;
 
     printf("[Status] Testing complete\n");
@@ -793,8 +846,8 @@ end_loop:
     }
 
     /* Print column headers */
-    fprintf(plotf, "# time target_qps actual_qps "
-                   "responses_per_sec failures_per_sec avg_latency\n");
+    fprintf(plotf, "# time target_qps actual_qps responses_per_sec failures_per_sec avg_latency"
+                   " connections conn_avg_latency\n");
 
     /* Don't print unused buckets */
     last_bucket_used = find_bucket(wait_phase_began) - buckets;
@@ -809,13 +862,18 @@ end_loop:
         double target_qps = t <= ramp_dtime ? (t / ramp_dtime) * max_qps : max_qps;
         double latency    = buckets[i].responses ? buckets[i].latency_sum / buckets[i].responses : 0;
         double interval   = bucket_interval / (double)MILLION;
-        fprintf(plotf, "%7.3f %8.2f %8.2f %8.2f %8.2f %8.6f\n",
+
+        double conn_latency = buckets[i].connections ? buckets[i].conn_latency_sum / buckets[i].connections : 0;
+
+        fprintf(plotf, "%7.3f %8.2f %8.2f %8.2f %8.2f %8.6f %8.2f %8.6f\n",
             t,
             target_qps,
-            buckets[i].queries / interval,
-            buckets[i].responses / interval,
-            buckets[i].failures / interval,
-            latency);
+            (double)buckets[i].queries / interval,
+            (double)buckets[i].responses / interval,
+            (double)buckets[i].failures / interval,
+            latency,
+            (double)buckets[i].connections / interval,
+            conn_latency);
     }
 
     fclose(plotf);
